﻿using ProtoBuf;
using ProtoBuf.Meta;
using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;

namespace PlayRealSerializerGenerator
{
    class Program
    {
        static void Main(string[] args)
        {
            GenerateRuntimeTypeModelAssembly(args);
        }

        private static void GenerateRuntimeTypeModelAssembly(string[] args)
        {
            //--- Interpret command line -i for input assembly (DTO)
            List<string> filesInput = new List<string>();
            string prevOpt = string.Empty;

            foreach (string arg in args)
            {
                if (arg == "-i")
                    prevOpt = arg;
                else if (prevOpt == "-i")
                {
#if DEBUG
                    filesInput.Add(Path.Combine(Environment.CurrentDirectory, arg));
#else
                    filesInput.Add(arg);
#endif
                }
            }
            //---

            if (filesInput.Count == 0)
            {
                Console.WriteLine("No input files specified");
                return;
            }

            List<Type> publicTypes = new List<Type>();

            foreach (string assemblyFileName in filesInput)
            {
                if (File.Exists(assemblyFileName))
                {
                    Assembly assembly = Assembly.LoadFrom(assemblyFileName);

                    foreach (Type type in assembly.GetTypes())
                    {
                        if (type.IsPublic)
                            publicTypes.Add(type);
                    }
                }
                else
                {
                    Console.WriteLine("File {0} not found", assemblyFileName);
                    return;
                }
            }


            RuntimeTypeModel GameDataModel = TypeModel.Create();

            //GameDataModel.AllowParseableTypes = true;
            GameDataModel.AutoCompile = false;
            GameDataModel.AutoAddMissingTypes = true;

            var subClasses = new List<Type>();

            foreach (var type in publicTypes)
            {
                List<Type> localSubClasses = new List<Type>();

                foreach (Type localType in publicTypes)
                {
                    if (localType.IsSubclassOf(type))
                        localSubClasses.Add(localType);
                }

                if (type.IsInterface)
                {
                    foreach (Type localType in publicTypes)
                    {
                        if (type.IsAssignableFrom(localType)
                            && localType != type
                            && !typeof(IExtensible).IsAssignableFrom(localType))
                        {
                            localSubClasses.Add(localType);
                        }
                    }
                }

                foreach (Type subClass in localSubClasses)
                {
                    if (!subClasses.Contains(subClass))
                    {
                        subClasses.Add(subClass);
                    }
                }
            }

            int cmp = 0;

            foreach (var type in publicTypes)
            {
                List<Type> localSubClasses = new List<Type>();

                foreach (Type localType in publicTypes)
                {
                    if (localType.IsSubclassOf(type))
                        localSubClasses.Add(localType);
                }

                if (type.IsInterface)
                {
                    foreach (Type localType in publicTypes)
                    {
                        if (type.IsAssignableFrom(localType)
                            && localType != type
                            && !typeof(IExtensible).IsAssignableFrom(localType))
                            localSubClasses.Add(localType);
                    }
                }

                if (!subClasses.Contains(type))
                {
                    Console.WriteLine("* {0}", type.Name);
                }

                cmp++;

                var meta = GameDataModel.Add(type, true);
                for (int i = 0; i < localSubClasses.Count; i++)
                {
                    Type subClass = localSubClasses[i];
                    Console.WriteLine("---> {0}", subClass.Name);

                    //GameDataModel.Add(subClass, true);
                    //NOTE: Think about backward compatibility in case of adding new subclasses
                    meta.AddSubType(10 + i, subClass);
                }
            }

            Console.WriteLine("Type count included : {0} / {1}", cmp, publicTypes.Count);

            GameDataModel.Compile("PlayRealSerializer", "PlayRealSerializer.dll");
        }
    }
}
